/*
 * swmod_opkg.c: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include "tools.h"
#include "swmod.h"
#include "swmod_uci.h"
#include "swmod_opkg.h"
#include "opkg_utils.h"
#include <curl/curl.h>
#include "swmod_api.h"

static char compulsory_service[][64] = {
	"opkg",
	"ubus",
	"ubusd",
	"procd",
	"swmodd"
};

int download_remote_package(const char *url, const char *username, const char *password, const char *dst)
{
	long res_code = 0;

	if (url == NULL || dst == NULL)
		return -1;

	CURL *curl = curl_easy_init();
	if (curl) {
		curl_easy_setopt(curl, CURLOPT_URL, url);

		if (username != NULL) {
			if (username[0] != '\0')
				curl_easy_setopt(curl, CURLOPT_USERNAME, username);
		}

		if (password != NULL) {
			if (password[0] != '\0')
				curl_easy_setopt(curl, CURLOPT_PASSWORD, password);
		}

		curl_easy_setopt(curl, CURLOPT_TIMEOUT, 100);

		FILE *fp = fopen(dst, "wb");
		if (fp) {
			curl_easy_setopt(curl, CURLOPT_WRITEDATA, fp);
			curl_easy_perform(curl);
			fclose(fp);
		} else {
			curl_easy_cleanup(curl);
			return -1;
		}

		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &res_code);
		curl_easy_cleanup(curl);

		if ((strncmp(url, "http", 4) == 0 && res_code != 200) ||
			(strncmp(url, "ftp", 3) == 0 && res_code != 226) ||
			(strncmp(url, "http", 4) && strncmp(url, "ftp", 3))) {
			return -1;
		}

		return 0;
	}

	return -1;
}

static bool validate_args(const char *url, char *output, int out_len)
{
	if (output == NULL || out_len <= 0) {
		return false;
	}

	if (url == NULL) {
		snprintf(output, out_len, "Package path not given");
		return false;
	}

	if (url[0] == '\0') {
		snprintf(output, out_len, "Package path not given");
		return false;
	}

	return true;
}

int swmod_perform_installation(const char *url, const char *uname, const char *pwd, char *output, int out_len)
{
	int err = -1;
	char package_path[2049] = {0};

	if (validate_args(url, output, out_len) == false)
		return err;

	if (strchr(url, ':') != NULL) {
		/* This is a remote package, need to download first */
		if(0 != download_remote_package(url, uname, pwd, DW_PKG_LOCAL_PATH)) {
			snprintf(output, out_len, "Package download failed from %s", url);
			return err;
		}

		snprintf(package_path, sizeof(package_path), "%s", DW_PKG_LOCAL_PATH);
	} else {
		snprintf(package_path, sizeof(package_path), "%s", url);
	}

	err = opkg_install_pkg(package_path, output, out_len);
	if (0 != err) {
		snprintf(output, out_len, "Internal Error");
	}

	/* Remove the downloaded package if it is in local dir of host or container */
	if (file_exists(DW_PKG_LOCAL_PATH)) {
		remove(DW_PKG_LOCAL_PATH);
	}

	return err;
}

int swmod_perform_upgrade(const char *url, const char *uname, const char *pwd, char *output, int out_len)
{
	int err = -1;
	char package_path[2049] = {0};

	if (validate_args(url, output, out_len) == false)
		return err;

	if (strchr(url, ':') != NULL) {
		/* This is a remote package, need to download first */
		if(0 != download_remote_package(url, uname, pwd, DW_PKG_LOCAL_PATH)) {
			snprintf(output, out_len, "Package download failed from %s", url);
			return err;
		}

		snprintf(package_path, sizeof(package_path), "%s", DW_PKG_LOCAL_PATH);
	} else {
		snprintf(package_path, sizeof(package_path), "%s", url);
	}

	err = opkg_update_pkg(package_path, output, out_len);
	if (0 != err) {
		snprintf(output, out_len, "Internal Error");
	}

	/* Remove the downloaded package if it is in local dir of host or container */
	if (file_exists(DW_PKG_LOCAL_PATH)) {
		remove(DW_PKG_LOCAL_PATH);
	}

	return err;
}

int swmod_perform_uninstall(const char *pname, char *output, int outlen)
{
	int err = -1;

	if (validate_args(pname, output, outlen) == false)
		return err;

	char version[MAX_LEN_64] = {0};
	char line[MAX_LEN_512] = {0};
	opkg_get_version(pname, version, sizeof(version));
	snprintf(line, sizeof(line), "\nPackage Name:%s Version:%s\n", pname, version);

	err = opkg_remove_pkg(pname, output, outlen);
	if (0 != err) {
		snprintf(output, outlen, "Internal Error");
	}

	int line_len = strlen(line);
	if (strlen(output) + line_len < outlen) {
		strncat(output, line, line_len);
	}

	return err;
}

static void get_description_from_package(struct uci_section *s, char *data)
{
	char *spch = NULL;

	if (s == NULL || data == NULL)
		return;

	if ((spch = strstr(data, "Description:"))) {
		remove_new_line(spch);
		swmod_uci_set_value_by_section(s, "description", spch+14);
	}
}

static void get_config_from_package(struct uci_section *s, struct list_head *opkg_files)
{
	struct buffer_list *iter;

	list_for_each_entry(iter, opkg_files, list) {
		char *spch = NULL;
		if ((spch = strstr(iter->line, STD_UCI_PATH))) {
			remove_new_line(spch);
			swmod_uci_set_value_by_section(s, "config", spch+12);
		}
		if ((spch = strstr(iter->line, STD_INIT_PATH))) {
			remove_new_line(spch);
			swmod_uci_set_value_by_section_list(s, "service", spch+12);
		}
	}

}

bool is_opkg_service(char *info)
{
	/* Get Config from package_name.list */
	if (strstr(info, STD_INIT_PATH) != NULL) {
		return true;
	}

	return false;
}

void populate_opkg_deployment_unit(ExecEnv_t *ee, struct list_head *pkg_list)
{
	char opkg_info[MAX_LEN_1024] = {0};
	struct uci_section *s = NULL, *stmp = NULL;
	struct buffer_list *bp;
	char env_id[4];

	if (ee->exists == false)
		return;

	if (swmod_uci_init(SWMOD_PATH))
		return;

	snprintf(env_id, 4, "%d", ee->eeid);

	swmod_uci_foreach_section_safe(ee->map_du_name, "deployment", stmp, s) {
		char *du_name = swmod_uci_get_value_by_section(s, "name");

		/* Check package name exists, if no => remove the section */
		if (present_in_buffer_list(pkg_list, du_name) == false) {
			PRINT_DEBUG("pkg %s deleted, delet pkg in uci", du_name);
			swmod_uci_delete_by_section(s, NULL, NULL);
		}
	}

	list_for_each_entry(bp, pkg_list, list) {
		char name[MAX_LEN_64] = {0};
		char version[MAX_LEN_64] = {0};

		/* Get name and version for each line */
		if (sscanf(bp->line, "%63s - %63s", name, version)) {

			if (is_compulsory_service(name) == true) {
				continue;
			}

			char cmd[MAX_LEN_1024];

			snprintf(cmd, sizeof(cmd), "opkg files %s 2>&1", name);
			swmod_run_cmd(ee, cmd, opkg_info, sizeof(opkg_info));

			if (is_opkg_service(opkg_info) == false) {
				continue;
			}

			/* Check if package name exists */
			struct uci_section *sec;
			if (check_section_exist_by_option(ee->map_du_name, "deployment", "name", name, &sec)) {
				const char *val = swmod_uci_get_value_by_section(sec, "version");
				if (strcmp(val, version) != 0) {
					swmod_uci_set_value_by_section(sec, "version", version);
				}
				continue;
			}

			LIST_HEAD(opkg_files_list);
			create_buffer_list(&opkg_files_list, opkg_info, sizeof(opkg_info));
			//dump_buffer_list(&opkg_files_list);

			struct uci_section *new_s = swmod_uci_add_section(ee->map_du_name, "deployment");
			static int incr = 0;

			/* Generate UUID */
			char *uuid = generate_uuid();

			/* Generate DUID */
			char *duid = generate_duid(true, incr++);

			swmod_uci_set_value_by_section(new_s, "name", name);
			swmod_uci_set_value_by_section(new_s, "version", version);
			swmod_uci_set_value_by_section(new_s, "uuid", uuid);
			swmod_uci_set_value_by_section(new_s, "duid", duid);
			swmod_uci_set_value_by_section(new_s, "environment", ee->name);
			swmod_uci_set_value_by_section(new_s, "eeid", env_id);

			/* Freed all allocated memory */
			FREE(uuid);
			FREE(duid);

			/* Get Config from package_name.list and Set it to map_du file */
			get_config_from_package(new_s, &opkg_files_list);
			delete_buffer_list(&opkg_files_list);

			snprintf(cmd, sizeof(cmd), "grep 'Description:' /usr/lib/opkg/info/%s.control", name);
			swmod_run_cmd(ee, cmd, opkg_info, sizeof(opkg_info));
			/* Get Description from package_name.control and Set it to map_du file */
			get_description_from_package(new_s, opkg_info);
		}
	}
	swmod_uci_fini(ee->map_du_name);
}

bool is_compulsory_service(const char *service)
{
	for (int i = 0; i < sizeof(compulsory_service)/sizeof(compulsory_service[0]); i++) {
		if (strcmp(compulsory_service[i], service) == 0) {
			return true;
		}
	}

	return false;
}
